#include <stdio.h>
#include "othello.h"

int aji(char* board, int status);

#define THINK_DEPTH_MAX 9
#define REST_FINAL 11

static void initialize(void);
static int aji_think(char* board);
static int aji_final(char* board);

typedef struct{
  int hand, point;
}HRESULT;

/***
  Main
****/

static int posp[2][16] = {
  { 001, 010, 021, 012, 030, 003, 023, 032,
      022, 013, 031, 011, 020, 002, 000, 033,},
  { 010, 001, 022, 003, 030, 032, 023,
      012, 021, 031, 013, 011, 002, 020, 000, 033,},
};

static int nSpace;
static int spaces[64];

int aji(char* board, int status)
{
  int* p;
  int i;
  
  if(initp(status))
    initialize();
  if(passp(status) || initp(status)) return 0;
  
  for(nSpace = 0, i = 0; i < 64; i++)
    if(board[i] == SPACE) nSpace++;
  
  p = (nSpace > 28) ? posp[0] : posp[1];
  for(nSpace = 0, i = 0; i < 16; i++){
    if(board[p[i] ^ 000] == SPACE) spaces[nSpace++] = p[i] ^ 000;
    if(board[p[i] ^ 007] == SPACE) spaces[nSpace++] = p[i] ^ 007;
    if(board[p[i] ^ 070] == SPACE) spaces[nSpace++] = p[i] ^ 070;
    if(board[p[i] ^ 077] == SPACE) spaces[nSpace++] = p[i] ^ 077;
  }
  
  if(nSpace <= REST_FINAL)
    return aji_final(board);
  return aji_think(board);
}

/***
  util
***/

static int count_score(char* board)
{
  int s, i;
  
  for(s = 0, i = 0; i < 64; i++)
    if(board[i] == MY_COLOR)
      s++;
    else if(board[i] != SPACE)
      s--;
  
  return s;
}

static int dir_step[4] = {1, 8, 9, 7};
static int check_lim[64][8]; /* [pos][dir] */

static void make_check_lim(void)
{
  int i, j;
  
  for(i = 0; i < 64; i += 8)
    for(j = i; j < i + 8; j++)
      check_lim[j][0] = i, check_lim[j][4] = i + 8;
  for(i = 0; i < 8; i++)
    for(j = i; j < 64; j += 8)
      check_lim[j][1] = i, check_lim[j][5] = 64;
  for(i = 0; i < 8; i++)
    for(j = i; j < ((8 - i) << 3); j += 9)
      check_lim[j][2] = i, check_lim[j][6] = ((8 - i) << 3);
  for(i = 8; i < 64; i += 8)
    for(j = i; j < 64; j += 9)
      check_lim[j][2] = i, check_lim[j][6] = 64;
  for(i = 0; i < 8; i++)
    for(j = i; j < (i << 3) + 1; j += 7)
      check_lim[j][3] = i, check_lim[j][7] = (i << 3) + 1;
  for(i = 15; i < 64; i += 8)
    for(j = i; j < 64; j += 7)
      check_lim[j][3] = i, check_lim[j][7] = 64;
}

static int chk_rev(char* board, int pos, int col)
{
  int i, l, n, m, p, s, x;

  if(board[pos] != SPACE)
    return 0;
  x = (col == MY_COLOR) ? OP_COLOR : MY_COLOR;
  n = 0;
  for(i = 0; i < 4; i++){
    l = check_lim[pos][i], s = dir_step[i], m = 0;
    for(p = pos - s; p >= l; p -= s)
      if(board[p] == x){
	m++;
      }else{
	if(board[p] == col)
	  n += m;
	break;
      }
  }      
  for(; i < 8; i++){
    l = check_lim[pos][i], s = dir_step[i - 4], m = 0;
    for(p = pos + s; p < l; p += s)
      if(board[p] == x){
	m++;
      }else{
	if(board[p] == col)
	  n += m;
	break;
      }
  }
  return n;
}

static int reverse(char* board, int pos, int col)
{
  int i, l, n, m, p, s, x;

  if(board[pos] != SPACE)
    return 0;
  x = (col == MY_COLOR) ? OP_COLOR : MY_COLOR;
  n = 0;
  for(i = 0; i < 4; i++){
    l = check_lim[pos][i], s = dir_step[i], m = 0;
    for(p = pos - s; p >= l; p -= s)
      if(board[p] == x){
	m++;
      }else{
	if(board[p] == col){
	  n += m;
	  for(; m > 0; m--)
	    board[p += s] = col;
	}
	break;
      }
  }      
  for(; i < 8; i++){
    l = check_lim[pos][i], s = dir_step[i - 4], m = 0;
    for(p = pos + s; p < l; p += s)
      if(board[p] == x){
	m++;
      }else{
	if(board[p] == col){
	  n += m;
	  for(; m > 0; m--)
	    board[p -= s] = col;
	}
	break;
      }
  }
  if(n > 0)
    board[pos] = col;
  return n;
}

/***
  aji_think
***/

#define SIDE_COST 400
#define DIAG_COST -100

static int eval_op_hand(char* board)
{
  int v = 0;
  
  if(chk_rev(board, 001, OP_COLOR)) v += SIDE_COST;
  if(chk_rev(board, 010, OP_COLOR)) v += SIDE_COST;
  if(chk_rev(board, 006, OP_COLOR)) v += SIDE_COST;
  if(chk_rev(board, 017, OP_COLOR)) v += SIDE_COST;
  if(chk_rev(board, 060, OP_COLOR)) v += SIDE_COST;
  if(chk_rev(board, 071, OP_COLOR)) v += SIDE_COST;
  if(chk_rev(board, 067, OP_COLOR)) v += SIDE_COST;
  if(chk_rev(board, 076, OP_COLOR)) v += SIDE_COST;
  
  if(chk_rev(board, 011, OP_COLOR)) v += DIAG_COST;
  if(chk_rev(board, 016, OP_COLOR)) v += DIAG_COST;
  if(chk_rev(board, 061, OP_COLOR)) v += DIAG_COST;
  if(chk_rev(board, 066, OP_COLOR)) v += DIAG_COST;

  return v;
}

/* SPACE = 0, MY_COLOR = 1, OP_COLOR = 2 */
#define CORNER_RISK 40
static int corner_val[3][3] = {
  {0, -100, 100}, {175, 200, 150}, {-175, -150, -200},
};

static int eval_corner(int* corner)
{
  int v;
  
  v = corner_val[corner[0]][corner[1]]
    + corner_val[corner[0]][corner[3]]
    + corner_val[corner[0]][corner[4]];
  
  if(corner[0] == SPACE){
    if(corner[1] == SPACE){
      if(corner[2] == MY_COLOR)
	v += CORNER_RISK;
      if(corner[5] == MY_COLOR)
	v += CORNER_RISK;
    }
    if(corner[3] == SPACE){
      if(corner[6] == MY_COLOR)
	v += CORNER_RISK;
      if(corner[7] == MY_COLOR)
	v += CORNER_RISK;
    }
    if(corner[2] == SPACE){
      if(corner[5] == MY_COLOR)
	v += CORNER_RISK;
      if(corner[7] == MY_COLOR)
	v += CORNER_RISK;
      if(corner[8] == MY_COLOR)
	v += CORNER_RISK;
    }
  }
  
  return v;
}

static int eval_board(char* board)
{
  int corner[9];
  int v;
  
  v = eval_op_hand(board);
  
  corner[0] = board[000]; corner[1] = board[001]; corner[2] = board[002];
  corner[3] = board[010]; corner[4] = board[011]; corner[5] = board[012];
  corner[6] = board[020]; corner[7] = board[021]; corner[8] = board[022];
  v += eval_corner(corner);
  
  corner[0] = board[007]; corner[1] = board[006]; corner[2] = board[005];
  corner[3] = board[017]; corner[4] = board[016]; corner[5] = board[015];
  corner[6] = board[027]; corner[7] = board[026]; corner[8] = board[025];
  v += eval_corner(corner);
  
  corner[0] = board[070]; corner[1] = board[071]; corner[2] = board[072];
  corner[3] = board[060]; corner[4] = board[061]; corner[5] = board[062];
  corner[6] = board[050]; corner[7] = board[051]; corner[8] = board[052];
  v += eval_corner(corner);
  
  corner[0] = board[077]; corner[1] = board[076]; corner[2] = board[075];
  corner[3] = board[067]; corner[4] = board[066]; corner[5] = board[065];
  corner[6] = board[057]; corner[7] = board[056]; corner[8] = board[055];
  v += eval_corner(corner);
  
  return v;
}

#define THINK_VAL_MAX 100000000
#define THINK_VAL_MIN -100000000
static HRESULT aji_think_my(char* board, int min);
static int aji_think_pf;
static int aji_think_depth;

static HRESULT aji_think_op(char* board, int max)
{
  HRESULT hr, hre;
  char bd[64];
  int i;
  
  hr.hand = -1, hr.point = THINK_VAL_MIN;
  
  aji_think_depth++;
  if(aji_think_depth > THINK_DEPTH_MAX){
    hr.point = eval_board(board);
    return hr;
  }
  
  memcpy(bd, board, sizeof(bd));
  for(i = 0; i < nSpace; i++)
    if(board[spaces[i]] == SPACE)
      if(reverse(bd, spaces[i], OP_COLOR)){
	hre = aji_think_my(bd, hr.point);
	memcpy(bd, board, sizeof(bd));
	if(hr.hand == -1 || hr.point < hre.point){
	  hr.hand = spaces[i];
	  hr.point = hre.point;
	  if(max <= hr.point) break;
	}
      }
  if(hr.hand == -1)
    if(aji_think_pf){
      aji_think_pf = 0;
      hr.point = (count_score(board) < 0) ? THINK_VAL_MIN : THINK_VAL_MIN;
    }else{
      --aji_think_depth;
      aji_think_pf = 1;
      hr = aji_think_my(board, THINK_VAL_MIN);
      return hr;
    }
  else
    aji_think_pf = 0;
  
  --aji_think_depth;
  return hr;
}

static HRESULT aji_think_my(char* board, int min)
{
  HRESULT hr, hre;
  char bd[64];
  int i;
  
  hr.hand = -1, hr.point = THINK_VAL_MAX;
  
  aji_think_depth++;
  
  memcpy(bd, board, sizeof(bd));
  for(i = 0; i < nSpace; i++)
    if(board[spaces[i]] == SPACE)
      if(reverse(bd, spaces[i], MY_COLOR)){
	hre = aji_think_op(bd, hr.point);
	memcpy(bd, board, sizeof(bd));
	if(hr.hand == -1 || hr.point > hre.point){
	  hr.hand = spaces[i];
	  hr.point = hre.point;
	  if(min >= hr.point)
	    break;
	}
      }
  if(hr.hand == -1)
    if(aji_think_pf){
      aji_think_pf = 0;
      hr.point = (count_score(board) < 0) ? THINK_VAL_MIN : THINK_VAL_MIN;
    }else{
      --aji_think_depth;
      aji_think_pf = 1;
      hr = aji_think_op(board, THINK_VAL_MAX);
      return hr;
    }
  else
    aji_think_pf = 0;
  
  --aji_think_depth;
  return hr;
}

static int aji_think(char* board)
{
  HRESULT hr;
  
  aji_think_pf = 0;
  aji_think_depth = 0;
  hr = aji_think_my(board, THINK_VAL_MIN);
  
  return hr.hand;
}

/***
  aji_final
***/

#define FINAL_VAL_MAX 65
#define FINAL_VAL_MIN -65
static HRESULT aji_final_my(char* board, int min);
static int aji_final_pf;

static HRESULT aji_final_op(char* board, int max)
{
  HRESULT hr, hre;
  char bd[64];
  int i;
  
  hr.hand = -1, hr.point = FINAL_VAL_MIN;
  
  memcpy(bd, board, sizeof(bd));
  for(i = 0; i < nSpace; i++)
    if(board[spaces[i]] == SPACE)
      if(reverse(bd, spaces[i], OP_COLOR)){
	hre = aji_final_my(bd, hr.point);
	memcpy(bd, board, sizeof(bd));
	if(hr.hand == -1 || hr.point < hre.point){
	  hr.hand = spaces[i];
	  hr.point = hre.point;
	  if(max <= hr.point) break;
	}
      }
  if(hr.hand == -1)
    if(aji_final_pf){
      aji_final_pf = 0;
      hr.point = count_score(board);
    }else{
      aji_final_pf = 1;
      hr = aji_final_my(board, FINAL_VAL_MIN);
    }
  else
    aji_final_pf = 0;
  
  return hr;
}

static HRESULT aji_final_my(char* board, int min)
{
  HRESULT hr, hre;
  char bd[64];
  int i;
  
  hr.hand = -1, hr.point = FINAL_VAL_MAX;
  
  memcpy(bd, board, sizeof(bd));
  for(i = 0; i < nSpace; i++)
    if(board[spaces[i]] == SPACE)
      if(reverse(bd, spaces[i], MY_COLOR)){
	hre = aji_final_op(bd, hr.point);
	memcpy(bd, board, sizeof(bd));
	if(hr.hand == -1 || hr.point > hre.point){
	  hr.hand = spaces[i];
	  hr.point = hre.point;
	  if(min >= hr.point)
	    break;
	}
      }
  if(hr.hand == -1)
    if(aji_final_pf){
      aji_final_pf = 0;
      hr.point = count_score(board);
    }else{
      aji_final_pf = 1;
      hr = aji_final_op(board, FINAL_VAL_MAX);
    }
  else
    aji_final_pf = 0;
  
  return hr;
}

static int aji_final(char* board)
{
  HRESULT hr;
  

  aji_final_pf = 0;
  hr = aji_final_my(board, FINAL_VAL_MIN);

  return hr.hand;
}

/***
  initialize
***/

static void initialize(void)
{
  make_check_lim();
}
