#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "othello.h"
#include "othelloX.h"

void InitGraph(void)
{
  int i;

  for(i = 0; i < 64; i++)
    cache[i] = SPACE;

  edge = 20;
  border = 4;
  box = 64;
  unit = (border * 9) + (box * 8);

  width = height = (edge * 2) + unit;

  display = XOpenDisplay(NULL);

  black = MyColor(display, "black");
  white = MyColor(display, "white");
  green = MyColor(display, "green4");
  gray = MyColor(display, "gray60");

  window = XCreateSimpleWindow(display, RootWindow(display, 0), 0, 0,
			       width, height, 0, 0, green);
  gc = XCreateGC(display, window, 0, NULL);
  XSelectInput(display, window, ButtonPressMask | ExposureMask);
  XMapWindow(display, window);
  XFlush(display);

  return;
}

void DrawBoard(Display *display, GC gc)
{
  int i;

  XSetForeground(display, gc, gray);
    for(i = 0; i < 10; i++){
    XFillRectangle(display, window, gc, (edge + (border + box) * i), edge, border, unit);
    XFillRectangle(display, window, gc, edge, (edge + (border + box) * i), unit, border);
  }
  XFlush(display);

  return;
}

unsigned long MyColor(Display *display, char *color)
{
  Colormap cmap;
  XColor c0, c1;

  cmap = DefaultColormap(display, 0);

  XAllocNamedColor(display, cmap, color, &c1, &c0);

  return(c1.pixel);
}

void DrawDisk(char *board)
{
  int x, y, pr, step, i;

  pr = edge + border + 4;
  step = box + border;

  for(i = 0; i < 64; i++){
    x = i % 8;
    y = i / 8;
    
    if(board[i] == WHITE){
      XSetForeground(display, gc, white);
      XFillArc(display, window, gc, (pr + x * step), (pr + y * step), 56, 56, 
	       0, 360*64);
    }else if(board[i] == BLACK){
      XSetForeground(display, gc, black);
      XFillArc(display, window, gc, (pr + x * step), (pr + y * step), 56, 56, 
	       0, 360*64);
    }
  }
  XFlush(display);
  
  return;
}

void CloseGraph(void)
{
  XDestroyWindow(display, window);
}
